#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2022
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
A demo implementation of a Helm-based DAliuGE deployment.

Limitations:
- Assumes graphs will run on a single pod
- Does not support external graph components (yet)
"""
import argparse
import json
import os
import logging
import tempfile

import dlg.exceptions
import dlg.restutils
from dlg.deploy.helm_client import HelmClient
from dlg.dropmake import pg_generator


def get_pg(opts, num_node_managers, num_data_island_managers):
    if not opts.logical_graph and not opts.physical_graph:
        return []

    if opts.logical_graph:
        unrolled_graph = pg_generator.unroll(opts.logical_graph)
        pgt = pg_generator.partition(
            unrolled_graph,
            algo="metis",
            num_partitions=num_node_managers,
            num_islands=num_data_island_managers,
        )
        del unrolled_graph
    else:
        with open(opts.physical_graph, "r", encoding="utf-8") as pg_file:
            pgt = json.load(pg_file)
    return pgt


def start_helm(physical_graph_template, num_nodes: int, deploy_dir: str):
    # TODO: Dynamic helm chart logging dir
    pgt = json.loads(physical_graph_template)
    pgt = pg_generator.partition(pgt, algo="metis", num_partitons=1, num_islands=1)
    helm_client = HelmClient(
        deploy_name="daliuge-daemon", chart_name="daliuge-daemon", deploy_dir=deploy_dir
    )
    helm_client.create_helm_chart(json.dumps(pgt), co_host=True)
    try:
        helm_client.launch_helm()
        helm_client.submit_pgt()
    except dlg.restutils.RestClientException as exp:
        raise exp
    except dlg.exceptions.InvalidGraphException as exp2:
        raise exp2
    finally:
        helm_client.teardown()
        logging.info("Finished deployment")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-L",
        "--logical-graph",
        action="store",
        type=str,
        dest="logical_graph",
        help="The filename of the logical graph to deploy",
        default=None,
    )
    parser.add_argument(
        "-P",
        "--physical_graph",
        action="store",
        type=str,
        dest="physical_graph",
        help="The filename of the physical graph (template) to deploy",
        default=None,
    )
    parser.add_argument(
        "-N",
        "--num_nodes",
        action="store",
        type=int,
        dest="num_nodes",
        help="The number of compute nodes you would like to try and deploy",
        default=1,
    )

    options = parser.parse_args()
    if bool(options.logical_graph) == bool(options.physical_graph):
        parser.error(
            "Either a logical graph or physical graph filename must be specified"
        )
    for graph_file_name in (options.logical_graph, options.physical_graph):
        if graph_file_name and not os.path.exists(graph_file_name):
            parser.error(f"Cannot locate graph_file at {graph_file_name}")

    if options.num_nodes <= 0:
        parser.error("The number of nodes must be a positive integer")

    physical_graph = get_pg(options, options.num_nodes, 1)
    # TODO: dynamic deployment directory.
    with tempfile.TemporaryDirectory() as tdir:
        helm_client = HelmClient(
            deploy_name="daliuge-daemon",
            chart_name="daliuge-daemon",
            deploy_dir=tdir,
        )
        helm_client.create_helm_chart(json.dumps(physical_graph))
        try:
            helm_client.launch_helm()
            helm_client.submit_and_monitor_pgt()
        except dlg.restutils.RestClientException as exp:
            raise exp
        except dlg.exceptions.InvalidGraphException as exp2:
            raise exp2
        finally:
            helm_client.teardown()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    main()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

import os
import subprocess
import sys
import sysconfig

from setuptools import find_packages
from setuptools import setup
from setuptools.command.install import install

# Version information
# We do like numpy: we have a major/minor/patch hand-written version written
# here. If we find the git commit (either via "git" command execution or in a
# dlg/version.py file) we append it to the VERSION later.
# The RELEASE flag allows us to create development versions properly supported
# by setuptools/pkg_resources or "final" versions.
MAJOR = 2
MINOR = 3
PATCH = 1
RELEASE = True
VERSION = "%d.%d.%d" % (MAJOR, MINOR, PATCH)
VERSION_FILE = "dlg/runtime/version.py"
PTH_FILE = "lib64_dist.pth"


def get_git_version():
    out = subprocess.check_output(["git", "rev-parse", "HEAD"])
    return out.strip().decode("ascii")


def get_version_info():
    git_version = "Unknown"
    if os.path.exists(".git"):
        git_version = get_git_version()
    full_version = VERSION
    if not RELEASE:
        full_version = "%s.dev0+%s" % (VERSION, git_version[:7])
    return full_version, git_version


def write_version_info():
    tpl = """
# THIS FILE IS GENERATED BY SETUP.PY
# DO NOT MODIFY BY HAND
version = '%(version)s'
git_version = '%(git_version)s'
full_version = '%(full_version)s'
is_release = %(is_release)s

if not is_release:
    version = full_version
"""
    full_version, git_version = get_version_info()
    with open(VERSION_FILE, "w") as f:
        info = tpl % {
            "version": VERSION,
            "full_version": full_version,
            "git_version": git_version,
            "is_release": RELEASE,
        }
        f.write(info.strip())


# Every time we overwrite the version file
write_version_info()


# HACK - HACK - HACK - HACK
class lib64_path(install):
    def write_pth_file(self):
        """
        This is a hack to get around some distributions installing stuff into
        lib64/python-x.x/dist-packages
        """
        lp = os.path.abspath(os.path.curdir)
        with open(PTH_FILE, "w") as f:
            f.write(lp.replace("/lib/", "/lib64/"))

    def initialize_options(self):
        install.initialize_options(self)
        self.build_scripts = None

    def finalize_options(self):
        install.finalize_options(self)
        self.set_undefined_options("build", ("build_scripts", "build_scripts"))

    def run(self):
        install.run(self)
        lp = sysconfig.get_path("stdlib")
        with open(PTH_FILE, "w") as f:
            f.write("{0}/dist-packages".format(lp))
        install.copy_file(self, PTH_FILE, os.path.join(self.install_lib, PTH_FILE))


# Core requirements of DALiuGE
# Keep alpha-sorted PLEASE!
install_requires = [
    "wheel",  # need to get wheel first...
    "bottle",
    "configobj",
    "crc32c",
    "daliuge-common==%s" % (VERSION,),
    "dill",
    "docker",
    "lockfile",
    # 0.10.6 builds correctly with old (<=3.10) Linux kernels
    "netifaces>=0.10.6",
    "numpy",
    "overrides",
    "paramiko",
    "psutil",
    "pyarrow",
    "python-daemon",
    "pyzmq ~= 22.3.0",
    "scp",
    "pyyaml",
    # 0.19.0 requires netifaces < 0.10.5, exactly the opposite of what *we* need
    "zeroconf >= 0.19.1",
    # 0.6 brings python3 support plus other fixes
    "zerorpc ~= 0.6.3",
]
# Keep alpha-sorted PLEASE!

# Extra requirements that are not needed by your every day daliuge installation
extra_requires = {
    # spead is required only for a specific app and its test, which we
    # skip anyway if spead is not found
    "spead": ["spead2==0.4.0"],
    # drive-casa is used by some manual tests under test/integrate
    "drive-casa": ["drive-casa>0.7"],
    # MPI support (MPIApp drops and HPC experiments) requires mpi4py
    "MPI": ["mpi4py"],
    # AWS storage types
    "aws": ["boto3"],
}

setup(
    name="daliuge-engine",
    version=get_version_info()[0],
    description=u"Data Activated \uF9CA (flow) Graph Engine - Execution Engine",
    long_description="""
        The element of the DALiuGE system executing the workflows. This replaces
        the former 'runtime' package (up to version 1.0). For more information 
        see the [Basics section(https://daliuge.readthedocs.io/en/latest/basics.html)]
        of the DALiuGE documentation.
        """,
    author="ICRAR DIA Group",
    author_email="dfms_prototype@googlegroups.com",
    url="https://github.com/ICRAR/daliuge",
    license="LGPLv2+",
    packages=find_packages(exclude=("test", "test.*", "fabfile")),
    package_data={
        "dlg.apps": ["dlg_app.h", "dlg_app2.h"],
        "dlg.manager": [
            "web/*.html",
            "web/static/css/*.css",
            "web/static/fonts/*",
            "web/static/js/*.js",
            "web/static/js/d3/*",
            "web/static/icons/*",
        ],
        "dlg.dropmake": [
            "web/lg_editor.html",
            "web/*.css",
            "web/*.js",
            "web/*.json",
            "web/*.map",
            "web/img/jsoneditor-icons.png",
            "web/pg_viewer.html",
            "web/matrix_vis.html",
            "lib/libmetis.*",
            "web/static/icons/*",
        ],
        "test.dropmake": ["logical_graphs/*.json"],
        "test.apps": ["dynlib_example.c", "dynlib_example2.c"],
    },
    entry_points={"dlg.tool_commands": ["runtime=dlg.runtime.tool_commands"]},
    install_requires=install_requires,
    extras_require=extra_requires,
    test_suite="test",
    cmdclass={"install": lib64_path},
)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
This package contains the modules implementing the core functionality of
the system.
"""

# It is of utmost importance that this is the *first* thing we do any time
# we load daliuge up. Most modules in daliuge that have a logger create it at
# import time, so we need to perform this setup before doing any other imports
def setup_logger_class():

    # Doc building is failing due with "ValueError: Empty module name" in
    # "from . import drop" further below, which I can't easily explain.
    # Just skip this setup in RTD for the time being
    import os

    if os.environ.get("READTHEDOCS", None) == "True":
        return

    import logging

    # To avoid 'No handlers could be found for logger' messages during testing
    logging.getLogger(__name__).addHandler(logging.NullHandler())

    # Use our own logger class, which knows about the currently executing app
    class _DlgLogger(logging.Logger):
        def makeRecord(self, *args, **kwargs):
            record = super(_DlgLogger, self).makeRecord(*args, **kwargs)

            from .. import drop
            from ..manager import session

            # Try to get the UID of the drop ultimately in charge of sending this
            # log record, if any
            try:
                drop = drop.track_current_drop.tlocal.drop
            except AttributeError:
                drop = None
            try:
                drop_uid = drop.uid if drop else ""
            except AttributeError:
                drop_uid = ""

            # Do the same with the session_id, which can be found via the drop (if any)
            # or checking if there is a session currently executing something
            session_id = ""
            if drop and hasattr(drop, "_dlg_session") and drop._dlg_session:
                session_id = drop._dlg_session.sessionId
            else:
                try:
                    session_id = session.track_current_session.tlocal.session.sessionId
                except AttributeError:
                    pass
            record.drop_uid = drop_uid
            record.session_id = session_id
            return record

    logging.setLoggerClass(_DlgLogger)


setup_logger_class()
del setup_logger_class

from dlg.common.version import git_version as __git_version__
from dlg.common.version import version as __version__
from ..apps import get_include_dir
from ..dask_emulation import delayed



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import os, string

__sub_tpl_str = """#!/bin/bash --login

#SBATCH --nodes=$NUM_NODES
#SBATCH --ntasks-per-node=1
#SBATCH --job-name=DALiuGE-$SESSION_ID
#SBATCH --time=$JOB_DURATION
#SBATCH --error=err-%j.log
$MODULES
$VENV

srun -l $PY_BIN -m dlg.deploy.start_dlg_cluster -l $LOG_DIR $GRAPH_PAR $PROXY_PAR $GRAPH_VIS_PAR $LOGV_PAR $ZERORUN_PAR $MAXTHREADS_PAR $SNC_PAR $NUM_ISLANDS_PAR $ALL_NICS $CHECK_WITH_SESSION --ssid $SESSION_ID --remote-mechanism slurm
"""
init_tpl = string.Template(__sub_tpl_str)


class DefaultConfig(object):
    MODULES = ""
    VENV = ""

    def __init__(self):
        self._dict = dict()
        l = self.init_list()
        self.setpar("acc", l[0])
        self.setpar("log_root", l[1])
        self.setpar("modules", l[2].strip())
        self.setpar("venv", l[3].strip())

    def init_list(self):
        pass

    def setpar(self, k, v):
        self._dict[k] = v

    def getpar(self, k):
        return self._dict.get(k)


#############################


class ICRARoodConfig(DefaultConfig):
    MODULES = """
    module load python/3.8.12
    """
    # The following is more a workaround than a solution
    # requires the user to have a venv exectly in that place
    ACCOUNT = os.environ["USER"]
    HOME_DIR = os.environ["HOME"]
    LOG_DIR = f"{HOME_DIR}/dlg/runs"
    VENV = f"source {HOME_DIR}/dlg/venv/bin/activate"

    def __init__(self):
        super(ICRARoodConfig, self).__init__()

    def init_list(self):  # TODO please fill in
        return [self.ACCOUNT, self.LOG_DIR, self.MODULES, self.VENV]


class GalaxyMWAConfig(DefaultConfig):
    def __init__(self):
        super(GalaxyMWAConfig, self).__init__()

    def init_list(self):
        return ["mwaops", "/group/mwaops/cwu/dfms/logs"]


class GalaxyASKAPConfig(DefaultConfig):
    MODULES = """
module swap PrgEnv-cray PrgEnv-gnu
module load python/2.7.10
module load mpi4py
"""
    VENV = ""

    def __init__(self):
        super(GalaxyASKAPConfig, self).__init__()

    def init_list(self):
        return ["astronomy856", "/group/astronomy856/cwu/dfms/logs", self.MODULES]


class MagnusConfig(DefaultConfig):
    def __init__(self):
        super(MagnusConfig, self).__init__()

    def init_list(self):
        return ["pawsey0129", "/group/pawsey0129/daliuge_logs"]


class TianHe2Config(DefaultConfig):
    def __init__(self):
        super(TianHe2Config, self).__init__()

    def init_list(self):  # TODO please fill in
        return ["SHAO", "/group/shao/daliuge_logs"]


##########################################


class ConfigFactory:
    mapping = {
        "galaxy_mwa": GalaxyMWAConfig,
        "galaxy_askap": GalaxyASKAPConfig,
        "magnus": MagnusConfig,
        "galaxy": GalaxyASKAPConfig,
        "shao": TianHe2Config,
        "hyades.icrar.org": ICRARoodConfig,
    }

    @staticmethod
    def available():
        return list(ConfigFactory.mapping.keys())

    @staticmethod
    def create_config(facility=None):
        facility = facility.lower() if (facility is not None) else facility
        return ConfigFactory.mapping.get(facility)()
